/*
 * Copyright 2005 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 *
 */

#include "../config.h"
#include <sys/param.h>
#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "buffer.h"

struct oddjob_buffer {
	unsigned char *data;
	size_t size, spare, used;
};

struct oddjob_buffer *
oddjob_buffer_new(size_t initial_size)
{
	struct oddjob_buffer *ret;
	ret = malloc(sizeof(struct oddjob_buffer));
	if (initial_size < 1024) {
		initial_size = 1024;
	}
	if (ret != NULL) {
		ret->data = malloc(initial_size);
		if (ret->data != NULL) {
			ret->size = initial_size;
			ret->spare = 0;
			ret->used = 0;
			ret->data[ret->spare + ret->used] = '\0';
		} else {
			free(ret);
			ret = NULL;
		}
	}
	return ret;
}

static void
oddjob_buffer_assure_avail(struct oddjob_buffer *buf, size_t minimum_size)
{
	unsigned char *data;
	size_t would_be_nice;
	if (buf->spare + buf->used + minimum_size > buf->size) {
		would_be_nice = (howmany(buf->spare + buf->used + minimum_size,
					 1024) + 1) * 1024;
		data = malloc(would_be_nice);
		if (data == NULL) {
			fprintf(stderr, "Out of memory\n");
			_exit(1);
		}
		memcpy(data, buf->data, buf->spare + buf->used);
		free(buf->data);
		buf->data = data;
		buf->size = would_be_nice;
	}
}

void
oddjob_buffer_prepend(struct oddjob_buffer *buf,
		      const unsigned char *bytes, size_t length)
{
	if (length == (size_t) -1) {
		length = strlen((const char *)bytes);
	}
	if (buf->spare > length) {
		memcpy(buf->data + buf->spare - length, bytes, length);
		buf->spare -= length;
	} else {
		oddjob_buffer_assure_avail(buf, length + 1);
		memmove(buf->data + buf->spare + length,
			buf->data + buf->spare, buf->used);
		memcpy(buf->data + buf->spare, bytes, length);
		buf->used += length;
	}
}

void
oddjob_buffer_append(struct oddjob_buffer *buf,
	      	     const unsigned char *bytes, size_t length)
{
	if (length == (size_t) -1) {
		length = strlen((const char *)bytes);
	}
	oddjob_buffer_assure_avail(buf, length + 1);
	memmove(buf->data + buf->spare + buf->used, bytes, length);
	buf->used += length;
	buf->data[buf->spare + buf->used] = '\0';
}

void
oddjob_buffer_consume(struct oddjob_buffer *buf, size_t length)
{
	if (length < buf->used) {
		buf->spare += length;
		buf->used -= length;
	} else {
		buf->used = 0;
	}
}

void
oddjob_buffer_clear(struct oddjob_buffer *buf)
{
	buf->used = 0;
}

void
oddjob_buffer_free(struct oddjob_buffer *buf)
{
	free(buf->data);
	buf->data = NULL;
	buf->size = 0;
	buf->used = 0;
	buf->spare = 0;
	free(buf);
}

size_t
oddjob_buffer_length(struct oddjob_buffer *buf)
{
	return buf->used;
}

const unsigned char *
oddjob_buffer_data(struct oddjob_buffer *buf)
{
	return buf->data + buf->spare;
}
